
//		UChars.cpp - functions for Unicode characters, including high-order characters


#define UNICODE							1
#define _UNICODE						1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR						1
#define VER_MINOR						4
#define VER_BUILD						1
#define VER_PATCH						1

#define PLUGIN_NAME						L"UChars"

#ifdef _WIN64
#define PLATFORM						L"x64"
#else
#define PLATFORM						L"x86"
#endif

#define MAX_ARG_LEN						8192
#define SHORT_BUF_LEN					1024
#define MAX_FILENAME_LEN				32768

#define ERRHANDLE						GetStdHandle( STD_ERROR_HANDLE )


#define SURROGATE_SIG_MASK				0xfc00
#define SURROGATE_SIG_HIGH				0xd800
#define SURROGATE_SIG_LOW				0xdc00
#define SURROGATE_MASK					0x03ff
#define SURROGATE_OFFSET				0x10000

#define LF								0x0a
#define CR								0x0d
#define QUOTE							0x22
#define NEL								0x85
#define BOM								0xfeff

#define CE_FLAG_SHOW_DECIMAL			0x0010
#define CE_FLAG_SHOW_CLASS				0x0020
#define CE_FLAG_SHOW_UTF16				0x0040
#define CE_FLAG_SHOW_UTF8				0x0080
#define CE_FLAG_SHOW_CHARACTER			0x0100
#define CE_FLAG_SHOW_NAME				0x0200

#define CHAR_NAME_MAX_LEN				64




const wchar_t PluginFeatures[] = L"@caps2,@lower2,@smiley,@uchar,@ucode,@ucodex,@ulen,@upper2,@ureverse,charencoding,toclip";



const wchar_t F_SmileyHelpText[]	= L"@SMILEY[value value...] \r\n\
\n\
Returns Unicode characters from the Emoticons block.\n\
\n\
Values must be in the range of 0 to 79.  Values may be decimal, or  \n\
hexadecimal with a leading 0x.  This function returns characters in  \n\
the range of U+1F600 to U+1F64F.  Separate values with spaces.\n";


const wchar_t F_UCharHelpText[]		= L"@UCHAR[value value...] \r\n\
\n\
Returns Unicode characters with the specified values.\n\
\n\
This function behaves like @CHAR, except that the input values\n\
are assumed to be hexadecimal, and characters outside the Basic  \n\
Multilingual Plane are supported.  You may prefix values with 0x  \n\
or U+ but neither is required.  With or without either prefix,\n\
each value will be parsed as hexadecimal.  Separate values with\n\
spaces.\n\
\n\
Values must be in the range of 1 to 10FFFF.\n";


const wchar_t F_UCodeHelpText[]		= L"@UCODE[string] \r\n\
\n\
Returns the hexadecimal values of characters in a string.\n\
\n\
This function behaves like @UNICODE, except that it returns  \n\
values as hexadecimal (without any prefix), and characters\n\
outside the Basic Multilingual Plane are supported.  A few\n\
characters, including the backquote and the closing square  \n\
bracket, will need to be escaped.\n";


const wchar_t F_UCodeXHelpText[]		= L"@UCODEX[string] \r\n\
\n\
Returns the hexadecimal values of characters in a string.\n\
\n\
This function behaves like @UNICODE, except that it returns  \n\
values as hexadecimal with a leading 0x, and characters\n\
outside the Basic Multilingual Plane are supported.  A few\n\
characters, including the backquote and the closing square  \n\
bracket, will need to be escaped.\n";


const wchar_t F_ULenHelpText[]		= L"@ULEN[string] \r\n\
\n\
Returns the number of Unicode characters in a string.\n\
\n\
This function is almost the same as @LEN, except that it counts  \n\
surrogate pairs as single characters.  (Surrogates which are not  \n\
properly paired will be counted as separate \u2018characters\u2019.)\n";


const wchar_t F_UReverseHelpText[]		= L"@UREVERSE[string] \r\n\
\n\
Returns string, with characters in reverse order.\n\
\n\
This function is almost the same as @REVERSE, except that it does  \n\
not mangle high-order characters by swapping surrogate characters.  \n";


const wchar_t F_Caps2HelpText[]			= L"@CAPS2[string] \r\n\
\n\
Returns string, with the first letter of each word  \n\
capitalized.  The default word separators are  \n\
whitespace characters and the hyphen.  You can  \n\
specify a different list of word separators via  \n\
an environment variable \u2018Caps2Separators\u2019.\n";


const wchar_t F_Lower2HelpText[]		= L"@LOWER2[string] \r\n\
\n\
Returns string, with all letters forced  \n\
to lowercase.\n";


const wchar_t F_Upper2HelpText[]		= L"@UPPER2[string] \r\n\
\n\
Returns string, with all letters forced  \n\
to uppercase.\n";


const wchar_t CharEncodingHelpText[]	= L"Show UTF-16 and UTF-8 encodings for characters.\n\
\n\
CHARENCODING /16 /8 /C /D /K /N /X value \"string\" ...\r\n\
\n\
\t/16\tUTF-16 encoding\n\
\t/8\tUTF-8 encoding\n\
\t/C\tcharacters\n\
\t/D\tdecimal values\n\
\t/K\tshow character class\n\
\t/N\tshow character name if available\n\
\t/X\texpand C-style escapes\n\
\n\
You may enter characters as quoted string literals, character\n\
values, HTML character entities, or any combination.  You may  \n\
prefix hex values with 0x or U+ but neither is required.  With  \n\
or without either prefix, hexadecimal is assumed.  Separate\n\
values with spaces.  If you specify neither /16 nor /8, the\n\
default is to show both.\n";


const wchar_t ToClipHelpText[]		= L"Copy the text on the command line to the clipboard.\n\
\n\
TOCLIP /A /H /Q /X text \r\n\
\n\
\t/A\tappend\n\
\t/H\texpand HTML entities\n\
\t/Q\tquietly\n\
\t/X\texpand C-style escapes\n\
\n\
The text on the command line will be put on the clipboard.\n\
Text may not begin with a slash.\n";



const wchar_t Caps2SepsVar[]		= L"Caps2Separators";



const int NumHelpTexts							= 11;
const wchar_t HelpFeature[NumHelpTexts][13]		= { L"@caps2", L"@lower2", L"@smiley", L"@uchar", L"@ucode", L"@ucodex", L"@ulen", L"@upper2", L"@ureverse",
													L"charencoding", L"toclip" };

const wchar_t *HelpText[NumHelpTexts]			= { F_Caps2HelpText, F_Lower2HelpText, F_SmileyHelpText, F_UCharHelpText, F_UCodeHelpText, F_UCodeXHelpText,
													F_ULenHelpText, F_Upper2HelpText, F_UReverseHelpText, CharEncodingHelpText, ToClipHelpText };



//		--------------------  Global Variables  --------------------

typedef unsigned __int32 ( __cdecl *_u_charNameProc ) ( __int32 code, int nameChoice, char *buffer, unsigned __int32 bufferLength, int *pErrorCode );

HMODULE UnicodeDll					= NULL;
_u_charNameProc u_charName			= NULL;


//		------------------------------------------------------------




int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetLastErrorText( LPTSTR outString )
{
	return GetErrorText( GetLastError(), outString );
}



int AppendFilename( LPTSTR Path, LPTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;


	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



HMODULE LoadUnicodeDll()
{
	wchar_t Filename[MAX_FILENAME_LEN]		= L"";

	size_t len								= 0;
	HMODULE LibHandle						= NULL;


	GetSystemDirectory( Filename, MAX_FILENAME_LEN );
	if ( Filename[0] )
		AppendFilename( Filename, L"", MAX_FILENAME_LEN );

	len = wcslen( Filename );

	wcscat_s( Filename, MAX_FILENAME_LEN, L"icu.dll" );
	if ( LibHandle = LoadLibrary( Filename ) )
		return LibHandle;

	Filename[len] = '\0';

	wcscat_s( Filename, MAX_FILENAME_LEN, L"icuuc.dll" );
	if ( LibHandle = LoadLibrary( Filename ) )
		return LibHandle;

	return NULL;
}



#include "ParseArgs.cpp"

#define EXPAND_ESCAPES_REAL_BACKSPACE
#include "ExpandEscapes.cpp"



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/uchars.html";
	piInfo.pszDescription	= (LPTSTR) L"Unicode character functions";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}


DLLExports BOOL WINAPI InitializePlugin( void )
{
	if ( UnicodeDll = LoadUnicodeDll() )
		u_charName = (_u_charNameProc) GetProcAddress( UnicodeDll, "u_charName" );


	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}


DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{

#ifdef NEWHELP_VERSION
	CloseHelpWindow();
#endif

	if ( UnicodeDll )
		FreeLibrary( UnicodeDll );

	return 0;
}



/*
int ReplaceChar( LPTSTR String, const wchar_t From, const wchar_t To )
{
	if ( !String || !From || !To )
		return -1;

	if ( From == To )
		return 0;


	int k			= 0;


	for ( wchar_t *ch = String; *ch; ch++ )
		if ( *ch == From )
			*ch = To, k++;

	return k;
}
*/



int StringToUnicode( LPTSTR String, int Format )
{
	if ( !String )
		return 1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	wchar_t val[SHORT_BUF_LEN]			= L"";
	size_t len							= 0;
	wchar_t *ch							= String;
	unsigned int CharVal				= 0;


	EscapeLine( String );

	while ( *ch ) {
		CharVal = *ch;

		if ( ( CharVal & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH )
			if ( ( ch[1] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW )
				CharVal = ( ( ( ch[0] & SURROGATE_MASK ) << 10 ) | ( ch[1] & SURROGATE_MASK ) ) + SURROGATE_OFFSET, ch++;

		switch ( Format ) {

		case 1 :
			wsprintf( val, L"0x%x ", CharVal );
			break;

		case 2 :
			wsprintf( val, L"U+%04X ", CharVal );
			break;

		case 3 :
			wsprintf( val, L"%u ", CharVal );
			break;

		default :
			wsprintf( val, L"%x ", CharVal );
		}

		if ( ( len + wcslen( val ) ) >= MAX_FILENAME_LEN ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
			return 2;
		}

		wcscat_s( temp, MAX_FILENAME_LEN, val );
		len += wcslen( val );
		ch++;
	}

	if ( len )
		temp[--len] = '\0';

	wcscpy_s( String, MAX_FILENAME_LEN, temp );
	return 0;
}



DLLExports int WINAPI f_ucode( LPTSTR String )
{
	return StringToUnicode( String, 0 );
}



DLLExports int WINAPI f_ucodex( LPTSTR String )
{
	return StringToUnicode( String, 1 );
}



DLLExports int WINAPI f_ulen( LPTSTR String )
{
	if ( !String )
		return 1;


	size_t n			= 0;
	wchar_t *ch			= String;


	while ( *ch ) {
		n++;

		if ( ( ( ch[0] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH ) && ( ( ch[1] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW ) )
			ch++;

		ch++;
	}

	wsprintf( String, L"%I64u", n );
	return 0;
}



DLLExports int WINAPI f_ureverse( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t t			= '\0';


	for ( wchar_t *ch = String; *ch; ch++ ) {
		if ( ( ( ch[0] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH ) && ( ( ch[1] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW ) )
			t = ch[0], ch[0] = ch[1], ch[1] = t, ch++;
	}

	_wcsrev( String );
	return 0;
}



DLLExports int WINAPI f_uchar( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	unsigned int j						= 0;
	wchar_t *ch							= String;
	wchar_t *sep1						= NULL;
	unsigned int value					= 0;
	unsigned int k						= 0;


	while ( *ch ) {

		while ( iswspace( *ch ) )
			ch++;

		if ( *ch == '\0' )
			break;

		if ( ( ch[0] == '0' ) && ( towupper( ch[1] ) == 'X' ) && iswxdigit( ch[2] ) )
			value = wcstoul( ch + 2, &sep1, 16 );
		else if ( ( towupper( ch[0] ) == 'U' ) && ( ch[1] == '+' ) && iswxdigit( ch[2] ) )
			value = wcstoul( ch + 2, &sep1, 16 );
		else if ( ( ch[0] == '#' ) && iswdigit( ch[1] ) )
			value = wcstoul( ch + 1, &sep1, 10 );
		else if ( iswxdigit( ch[0] ) )
			value = wcstoul( ch, &sep1, 16 );
		else
			goto Fail;

		if ( *sep1 && !iswspace( *sep1 ) )
			goto Fail;

		if ( ( value == 0 ) || ( value > 0x10ffff ) )
			goto Fail2;

		ch = sep1;

		k = ( value > 0xffff ) ? 2 : 1;

		if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
			return -1;
		}

		if ( k == 2 ) {
			temp[j++]	= ( ( value - SURROGATE_OFFSET ) >> 10 ) | SURROGATE_SIG_HIGH;
			temp[j++]	= ( ( value - SURROGATE_OFFSET ) & SURROGATE_MASK ) | SURROGATE_SIG_LOW;
			temp[j]		= '\0';
		}
		else {
			temp[j++]	= value;
			temp[j]		= '\0';
		}
	}


	wcscpy_s( String, MAX_ARG_LEN, temp );
	return 0;


Fail :

	sep1 = ch;
	while ( *sep1 && !iswspace( *sep1 ) )
		sep1++;

	if ( *sep1 )
		*sep1 = '\0';

	_wcsupr_s( ch, MAX_ARG_LEN );
	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Bad character value: \"%s\"\r\n", ch );
	return -1;

Fail2 :

	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Bad character value; should be 0x0001 - 0x10FFFF: 0x%04X\r\n", value );
	return -1;
}



DLLExports int WINAPI f_smiley( LPTSTR String )
{
	if ( !String )
		return 0;


	wchar_t temp[MAX_ARG_LEN]		= L"";

	wchar_t *sep1			= NULL;
	unsigned int val		= 0;
	unsigned int j			= 0;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			val = 0;

			if ( ( arg[i][0] == '0' ) && ( towupper( arg[i][1] ) == 'X' ) && iswxdigit( arg[i][2] ) )
				val = wcstoul( arg[i] + 2, &sep1, 16 );
			else if ( iswdigit( arg[i][0] ) )
				val = wcstoul( arg[i], &sep1, 10 );
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] );
				return -1;
			}

			if ( val >= 80 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Value must be 0 - 79: %u\r\n", val );
				return -1;
			}

			if ( j < ( MAX_ARG_LEN - 2 ) ) {
				val += 0xf600;
				temp[j++]	= ( val >> 10 ) | SURROGATE_SIG_HIGH;
				temp[j++]	= ( val & SURROGATE_MASK ) | SURROGATE_SIG_LOW;
				temp[j]		= '\0';
			}
		}
	}


	wcscpy_s( String, MAX_ARG_LEN, temp );
	return 0;
}



DLLExports int WINAPI f_upper2( LPTSTR String )
{
	if ( !String )
		return 1;


	CharUpperW( String );
	return 0;
}



DLLExports int WINAPI f_lower2( LPTSTR String )
{
	if ( !String )
		return 1;


	CharLowerW( String );
	return 0;
}



DLLExports int WINAPI f_caps2( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t Separators[SHORT_BUF_LEN]			= L"";

	BOOL InWord			= FALSE;


	GetEnvironmentVariable( Caps2SepsVar, Separators, SHORT_BUF_LEN );


	for ( wchar_t *ch = String; *ch; ch++ ) {

		if ( IsCharAlphaW( *ch ) )
			if ( !InWord )
				*ch = (wchar_t) CharUpperW( (LPWSTR) *ch );

		if ( Separators[0] ) {
			if ( wcschr( Separators, *ch ) )
				InWord = FALSE;
			else
				InWord = TRUE;
		}
		else {
			if ( iswspace( *ch ) || ( *ch == '-' ) )
				InWord = FALSE;
			else
				InWord = TRUE;
		}
	}

	return 0;
}



#include "HtmlEntities.cpp"



BOOL Noncharacter( unsigned int ch )
{
	if ( ( ch >= 0xfdd0 ) && ( ch <= 0xfdef ) )
		return TRUE;

	if ( ( ch & 0xffff ) >= 0xfffe )
		return TRUE;

	return FALSE;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI charencoding( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t temp[MAX_ARG_LEN]				= L"";
	wchar_t tmp2[MAX_ARG_LEN]				= L"";
	char CharName[CHAR_NAME_MAX_LEN]		= "";
	wchar_t CharNameW[CHAR_NAME_MAX_LEN]	= L"";

	wchar_t *sep1			= NULL;
	unsigned int val		= 0;
	size_t elen				= 0;
	size_t j				= 0;
	unsigned int k			= 0;
	int ErrorCode			= 0;

	BOOL Expand				= FALSE;
	unsigned int Flags		= 0;
	
	wchar_t Class			= '\0';
	BYTE b1					= 0;
	BYTE b2					= 0;
	BYTE b3					= 0;
	BYTE b4					= 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_RETAIN_QUOTES );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( CharEncodingHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( CharEncodingHelpText, TRUE );
			else if ( !wcscmp( arg[i], L"/16" ) )
				Flags |= CE_FLAG_SHOW_UTF16;
			else if ( !wcscmp( arg[i], L"/8" ) )
				Flags |= CE_FLAG_SHOW_UTF8;
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Flags |= CE_FLAG_SHOW_CHARACTER;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				Flags |= CE_FLAG_SHOW_DECIMAL;
			else if ( !_wcsicmp( arg[i], L"/K" ) )
				Flags |= CE_FLAG_SHOW_CLASS;
			else if ( !_wcsicmp( arg[i], L"/N" ) ) {
				if ( u_charName )
					Flags |= CE_FLAG_SHOW_NAME;
			}
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int j = 2;

				if ( !_wcsnicmp( arg[i] + j, L"TF", 2 ) )
					j += 2;

				if ( arg[i][j] == '-' )
					j++;

				if ( !wcscmp( arg[i] + j, L"16" ) )
					Flags |= CE_FLAG_SHOW_UTF16;
				else if ( !wcscmp( arg[i] + j, L"8" ) )
					Flags |= CE_FLAG_SHOW_UTF8;
				else {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
					return 0;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				Expand = TRUE;
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( arg[i][0] == QUOTE ) {
				wcscpy_s( tmp2, MAX_ARG_LEN, arg[i] + 1 );

				j = wcslen( tmp2 );
				if ( j )
					if ( tmp2[j - 1] == QUOTE )
						tmp2[--j] = '\0';

				if ( tmp2[0] ) {
					if ( Expand )
						ExpandEscapes( tmp2, MAX_ARG_LEN );

					wcscat_s( temp, MAX_ARG_LEN, tmp2 );
				}
			}
			else {
				val		= 0;
				sep1	= NULL;

				if ( !_wcsicmp( arg[i], L"BOM" ) )
					val = BOM;
				else if ( ( arg[i][0] == '0' ) && ( towupper( arg[i][1] ) == 'X' ) && iswxdigit( arg[i][2] ) )
					val = wcstoul( arg[i] + 2, &sep1, 16 );
				else if ( ( towupper( arg[i][0] ) == 'U' ) && ( arg[i][1] == '+' ) && iswxdigit( arg[i][2] ) )
					val = wcstoul( arg[i] + 2, &sep1, 16 );
				else if ( ( arg[i][0] == '#' ) && iswdigit( arg[i][1] ) )
					val = wcstoul( arg[i] + 1, &sep1, 10 );
				else if ( iswxdigit( arg[i][0] ) )
					val = wcstoul( arg[i], &sep1, 16 );

#ifdef NUM_HTML_ENTITIES

				else if ( arg[i][0] == '&' ) {
					val = ParseEntity( arg[i], &elen );

					if ( !val ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Unknown entity: \"%s\"\r\n", arg[i] );
						return 1;
					}
				}

#endif	//	NUM_HTML_ENTITIES

				else if ( ( arg[i][0] > 0 ) && ( ( arg[i][0] & SURROGATE_SIG_MASK ) != SURROGATE_SIG_HIGH ) && ( ( arg[i][0] & SURROGATE_SIG_MASK ) != SURROGATE_SIG_LOW ) && ( arg[i][1] == '\0' ) )
					val = arg[i][0];
				else if ( ( ( arg[i][0] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH ) && ( ( arg[i][1] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW ) && ( arg[i][2] == '\0' ) )
					val = ( ( ( arg[i][0] & SURROGATE_MASK ) << 10 ) | ( arg[i][1] & SURROGATE_MASK ) ) + SURROGATE_OFFSET;
				else {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown character: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( sep1 ) {
					if ( *sep1 ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad character value: \"%s\"\r\n", arg[i] );
						return 1;
					}
				}

				if ( ( val == 0 ) || ( val > 0x10ffff ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad character value; should be 0x0001 - 0x10FFFF: 0x%04X\r\n", val );
					return 1;
				}

				k = ( val > 0xffff ) ? 2 : 1;

				if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
					return 1;
				}

				if ( k == 2 ) {
					temp[j++]	= ( ( val - SURROGATE_OFFSET ) >> 10 ) | SURROGATE_SIG_HIGH;
					temp[j++]	= ( ( val - SURROGATE_OFFSET ) & SURROGATE_MASK ) | SURROGATE_SIG_LOW;
					temp[j]		= '\0';
					//	wprintf( L"   * Surrogates:  %04X  %04X\n", temp[j - 2], temp[j - 1] );
				}
				else {
					temp[j++]	= val;
					temp[j]		= '\0';
				}
			}
		}
	}


	if ( ( Flags & ( CE_FLAG_SHOW_UTF16 | CE_FLAG_SHOW_UTF8 ) ) == 0 )
		Flags |= ( CE_FLAG_SHOW_UTF16 | CE_FLAG_SHOW_UTF8 );


	//		Show header :

	CrLf();

	QPuts( L"Unicode   " );

	if ( Flags & CE_FLAG_SHOW_DECIMAL )
		QPuts( L"Decimal  " );

	if ( Flags & CE_FLAG_SHOW_CLASS )
		QPuts( L"K  " );

	if ( Flags & CE_FLAG_SHOW_UTF16 )
		QPuts( L"UTF-16     " );

	if ( Flags & CE_FLAG_SHOW_UTF8 )
		QPuts( L"UTF-8        " );

	if ( Flags & CE_FLAG_SHOW_CHARACTER )
		QPuts( L"Ch  " );

	if ( Flags & CE_FLAG_SHOW_NAME )
		QPuts( L"Name                                      " );

	QPuts( L"\r\n" );

	QPuts( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015  " );

	if ( Flags & CE_FLAG_SHOW_DECIMAL )
		QPuts( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015  " );

	if ( Flags & CE_FLAG_SHOW_CLASS )
		QPuts( L"\u2015  " );

	if ( Flags & CE_FLAG_SHOW_UTF16 )
		QPuts( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015  " );

	if ( Flags & CE_FLAG_SHOW_UTF8 )
		QPuts( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015  " );

	if ( Flags & CE_FLAG_SHOW_CHARACTER )
		QPuts( L"\u2015\u2015  " );

	if ( Flags & CE_FLAG_SHOW_NAME )
		QPuts( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015  " );

	CrLf();


	//		Iterate through string :

	for ( size_t i = 0; temp[i]; i++ ) {

		val			= temp[i];
		Class		= '-';

		if ( ( ( val & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH ) && ( ( temp[i + 1] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW ) ) {
			val = ( ( ( val & SURROGATE_MASK ) << 10 ) | ( temp[i + 1] & SURROGATE_MASK ) ) + SURROGATE_OFFSET;
			i++;
		}

		if ( val == 0xfeff )
			Class = 'B';
		else if ( Noncharacter( val ) )
			Class = 'N';
		else if ( ( val < 0x10000 ) && ( ( val & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH ) )
			Class = 'H';
		else if ( ( val < 0x10000 ) && ( ( val & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW ) )
			Class = 'L';
		else if ( iswcntrl( val ) )
			Class = 'C';
		else if ( iswspace( val ) )
			Class = 'W';
		else if ( iswalpha( val ) )
			Class = 'A';
		else if ( iswdigit( val ) )
			Class = 'D';
		else if ( iswpunct( val ) )
			Class = 'P';


		Printf( L"U+%06X  ", val );

		if ( Flags & CE_FLAG_SHOW_DECIMAL )
			Printf( L"%7u  ", val );

		if ( Flags & CE_FLAG_SHOW_CLASS )
			Printf( L"%c  ", Class );

		if ( Flags & CE_FLAG_SHOW_UTF16 ) {
			if ( val > 0xffff )
				Printf( L"%04X %04X  ", ( ( val - SURROGATE_OFFSET ) >> 10 ) | SURROGATE_SIG_HIGH, ( ( val - SURROGATE_OFFSET ) & SURROGATE_MASK ) | SURROGATE_SIG_LOW );
			else
				Printf( L"%04X       ", val );
		}

		if ( Flags & CE_FLAG_SHOW_UTF8 ) {
			if ( val < 0x80 )
				Printf( L"%02X           ", val );
			else if ( val < 0x800 ) {
				b1 = ( val >> 6 ) | 0xc0;
				b2 = ( val & 0x3f ) | 0x80;
				Printf( L"%02X %02X        ", b1, b2 );
			}
			else if ( val < 0x10000 ) {
				b1 = ( val >> 12 ) | 0xe0;
				b2 = ( ( val >> 6 ) & 0x3f ) | 0x80;
				b3 = ( val & 0x3f ) | 0x80;
				Printf( L"%02X %02X %02X     ", b1, b2, b3 );
			}
			else {
				b1 = ( val >> 18 ) | 0xf0;
				b2 = ( ( val >> 12 ) & 0x3f ) | 0x80;
				b3 = ( ( val >> 6 ) & 0x3f ) | 0x80;
				b4 = ( val & 0x3f ) | 0x80;
				Printf( L"%02X %02X %02X %02X  ", b1, b2, b3, b4 );
			}
		}

		if ( Flags & CE_FLAG_SHOW_CHARACTER ) {
			if ( wcschr( L"-ADPW", Class ) ) {
				if ( val <= 0xffff )
					Printf( L"%c   ", val );
				else
					Printf( L"%c%c  ", ( ( val - SURROGATE_OFFSET ) >> 10 ) | SURROGATE_SIG_HIGH, ( ( val - SURROGATE_OFFSET ) & SURROGATE_MASK ) | SURROGATE_SIG_LOW );
			}
			else
				QPuts( L"    " );
		}

		if ( ( Flags & CE_FLAG_SHOW_NAME ) && u_charName ) {
			CharNameW[0] = CharName[0] = '\0';
			u_charName( val, 0, CharName, CHAR_NAME_MAX_LEN, &ErrorCode );

			if ( CharName[0] ) {
				if ( wcslen( CharNameW ) > 40 ) {
					CharNameW[39] = '\u2026';
					CharNameW[40] = '\0';
				}

				MultiByteToWideChar( CP_OEMCP, 0, CharName, -1, CharNameW, CHAR_NAME_MAX_LEN );
				if ( CharNameW[0] )
					Printf( L"%-40.40s  ", CharNameW );
			}
		}

		CrLf();
	}

	return 0;
}



size_t CountEOLs( LPCTSTR String )
{
	if ( !String )
		return 0;


	size_t k				= 0;


	for ( const wchar_t *ch = String; *ch; ch++ )
		if ( ( *ch == CR ) || ( *ch == LF ) || ( *ch == NEL ) )
			k++;

	return k;
}



int ClipCopyText( LPTSTR Source, LPTSTR Dest, size_t bufsize )
{
	if ( !Source || !Dest || !bufsize )
		return -1;


	size_t j		= wcslen( Dest );
	int k			= 0;


	if ( j >= ( bufsize - 1 ) )
		return -1;

	for ( wchar_t *ch = Source; *ch; ch++ ) {

		if ( ( *ch == BOM ) && ( ch == Source ) )
			;
		else if ( ( *ch == CR ) || ( *ch == LF ) || ( *ch == NEL ) ) {

			if ( j < ( bufsize - 2 ) ) {
				Dest[j++]	= CR;
				Dest[j++]	= LF;
				Dest[j]		= '\0';
				k++;
			}

			if ( ( *ch == CR ) || ( *ch == LF ) )
				if ( ch[1] == ( *ch ^ 0x07 ) )
					ch++;
		}
		else if ( j < ( bufsize - 1 ) ) {
			Dest[j++]	= *ch;
			Dest[j]		= '\0';
		}
	}

	return 0;
}



size_t NewLen( LPCTSTR String )
{
	if ( !String )
		return 0;


	size_t k			= 0;


	for ( const wchar_t *ch = String; *ch; ch++ ) {
		k++;

		if ( ( *ch & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH )
			if ( ( ch[1] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW )
				ch++;
	}

	return k;
}



DLLExports int WINAPI toclip( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t TextToCopy[MAX_ARG_LEN]		= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";

	BOOL Append					= FALSE;
	BOOL Quiet					= FALSE;
	BOOL Expand					= FALSE;
	BOOL ExpandEntities			= FALSE;
	LPTSTR InputText			= NULL;
	LPTSTR OldClipboardText		= NULL;
	size_t NewTextSize			= 0;
	HGLOBAL NewTextHandle		= NULL;
	LPTSTR NewTextBuffer		= NULL;
	size_t newlen				= 0;

	/*
	const wchar_t invchar1		= 0xfdea;
	const wchar_t invchar2		= 0xfdeb;
	*/

	int rv				= 2;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_RETAIN_QUOTES | PARSE_ONE_ARG_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {

		if ( !wcscmp( arg[i], L"/?" ) )
			return ShowCmdHelp( ToClipHelpText, FALSE );
		else if ( !wcscmp( arg[i], L"/??" ) )
			return ShowCmdHelp( ToClipHelpText, TRUE );
		else if ( !_wcsicmp( arg[i], L"/A" ) )
			Append = TRUE;
		else if ( !_wcsicmp( arg[i], L"/H" ) )
			ExpandEntities = TRUE;
		else if ( !_wcsicmp( arg[i], L"/Q" ) )
			Quiet = TRUE;
		else if ( !_wcsicmp( arg[i], L"/X" ) )
			Expand = TRUE;
		else if ( arg[i][0] == '/' ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else
			InputText = arg[i];
	}


	if ( !InputText ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No text\r\n" );
		return 1;
	}


	wcscpy_s( TextToCopy, MAX_ARG_LEN, InputText );


	/*
	if ( Expand ) {
		if ( ExpandEntities )
			ReplaceChar( TextToCopy, '&', invchar1 );

		ExpandEscapes( TextToCopy, MAX_ARG_LEN );

		if ( ExpandEntities ) {
			ReplaceChar( TextToCopy, '&', invchar2 );
			ReplaceChar( TextToCopy, invchar1, '&' );
		}
	}

	if ( ExpandEntities ) {
		Deentitize( TextToCopy, MAX_ARG_LEN );

		if ( Expand )
			ReplaceChar( TextToCopy, invchar2, '&' );
	}
	*/


	if ( Expand )
		ExpandEscapes( TextToCopy, MAX_ARG_LEN );

	if ( ExpandEntities )
		Deentitize( TextToCopy, MAX_ARG_LEN );


	NewTextSize = wcslen( TextToCopy ) + CountEOLs( TextToCopy ) + 16;


	if ( OpenClipboard( GetConsoleWindow() ) ) {

		if ( Append ) {
			OldClipboardText = (LPTSTR) GetClipboardData( CF_UNICODETEXT );

			if ( OldClipboardText )
				NewTextSize = NewTextSize + wcslen( OldClipboardText ) + CountEOLs( OldClipboardText );
		}
		else
			OldClipboardText = NULL;

		NewTextHandle = GlobalAlloc( GMEM_MOVEABLE, NewTextSize * sizeof( wchar_t ) );

		if ( NewTextHandle ) {
			NewTextBuffer = (LPTSTR) GlobalLock( NewTextHandle );

			if ( NewTextBuffer ) {
				NewTextBuffer[0] = '\0';

				if ( OldClipboardText )
					ClipCopyText( OldClipboardText, NewTextBuffer, NewTextSize );

				ClipCopyText( TextToCopy, NewTextBuffer, NewTextSize );
				newlen = NewLen( NewTextBuffer );

				if ( SetClipboardData( CF_UNICODETEXT, NewTextBuffer ) ) {
					rv = 0;

					if ( !Quiet )
						Printf( L"Wrote %I64u character%s to the clipboard.\r\n", newlen, ( newlen == 1 ) ? L"" : L"s" );
				}
				else {
					GlobalUnlock( NewTextHandle );
					GetLastErrorText( ErrorText );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Error setting clipboard data:\r\n" );
					Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
				}
			}
			else {
				GetLastErrorText( ErrorText );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error locking global memory:\r\n" );
				Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
			}

			if ( rv )
				GlobalFree( NewTextHandle );
		}
		else {
			GetLastErrorText( ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error allocating global memory:\r\n" );
			Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
		}

		CloseClipboard();
	}
	else {
		GetLastErrorText( ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening clipboard:\r\n" );
		Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
	}

	return rv;
}



#ifndef	NEWHELP_VERSION

LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}

#endif	//	NEWHELP_VERSION



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}



#ifndef NEWHELP_VERSION

//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}

#endif		//	NEWHELP_VERSION



/*

	1.0.0.0		2021-07-19

	1.1.0.0		2022-06-02
				Adds @UREVERSE.  Also, included a newer HtmlEntities.cpp
				with support for all HTML 4 character entities.

	1.2.0.0		2022-11-16
				Added /X to both CHARENCODING and TOCLIP to support C-style
				character escapes.

	1.3.0.0		2022-11-16
				Added /H to TOCLIP to expand any HTML entities in the text.

	1.4.0.0		2023-04-07
				Added @CAPS2, @LOWER2, and @UPPER2.

	1.4.0.1		2023-07-24
				Updated ParseArgs.cpp to the current version.

	1.4.0.2		2023-07-31

	1.4.0.3		2023-10-12
				Updated the plugin's web address.

	1.4.1.0		2024-03-15
				Added /N to CHARENCODING to look up character names.

	1.4.1.1		2024-03-15
				Fixed the above to correctly handle characters outside the BMP.

*/
